
#include <postgres.h>
#include <fmgr.h>
#include <access/skey.h>
#include <access/gist.h>

#include "row_acl.h"
#include "row_acl_bits.h"
#include "row_acl_text.h"

/*
 * #ifdef PG_MODULE_MAGIC
 * PG_MODULE_MAGIC;
 * #endif/ * ? PG_MODULE_MAGIC * /
 */

/*
 * pg infos
 */
PG_FUNCTION_INFO_V1(row_acl_in);
PG_FUNCTION_INFO_V1(row_acl_out);

PG_FUNCTION_INFO_V1(row_acl_eq);
PG_FUNCTION_INFO_V1(row_acl_neq);
PG_FUNCTION_INFO_V1(row_acl_is_intersects);
PG_FUNCTION_INFO_V1(row_acl_contains);
PG_FUNCTION_INFO_V1(row_acl_within);

PG_FUNCTION_INFO_V1(int_to_row_acl);
PG_FUNCTION_INFO_V1(row_acl_to_int);
PG_FUNCTION_INFO_V1(text_to_row_acl);
PG_FUNCTION_INFO_V1(row_acl_to_text);

PG_FUNCTION_INFO_V1(row_acl_intersect);
PG_FUNCTION_INFO_V1(row_acl_union);
PG_FUNCTION_INFO_V1(row_acl_subtract);

PG_FUNCTION_INFO_V1(row_acl_set);
PG_FUNCTION_INFO_V1(row_acl_set_alt);

PG_FUNCTION_INFO_V1(gist_row_acl_compress);
PG_FUNCTION_INFO_V1(gist_row_acl_decompress);
PG_FUNCTION_INFO_V1(gist_row_acl_same);
PG_FUNCTION_INFO_V1(gist_row_acl_penalty);
PG_FUNCTION_INFO_V1(gist_row_acl_union);
PG_FUNCTION_INFO_V1(gist_row_acl_consistent);
PG_FUNCTION_INFO_V1(gist_row_acl_picksplit);

/*
 * prototypes
 */
Datum row_acl_in(PG_FUNCTION_ARGS);
Datum row_acl_out(PG_FUNCTION_ARGS);

Datum row_acl_eq(PG_FUNCTION_ARGS);
Datum row_acl_neq(PG_FUNCTION_ARGS);
Datum row_acl_is_intersects(PG_FUNCTION_ARGS);
Datum row_acl_contains(PG_FUNCTION_ARGS);
Datum row_acl_within(PG_FUNCTION_ARGS);

Datum int_to_row_acl(PG_FUNCTION_ARGS);
Datum row_acl_to_int(PG_FUNCTION_ARGS);
Datum text_to_row_acl(PG_FUNCTION_ARGS);
Datum row_acl_to_text(PG_FUNCTION_ARGS);

Datum row_acl_intersect(PG_FUNCTION_ARGS);
Datum row_acl_union(PG_FUNCTION_ARGS);
Datum row_acl_subtract(PG_FUNCTION_ARGS);

Datum row_acl_set(PG_FUNCTION_ARGS);
Datum row_acl_set_alt(PG_FUNCTION_ARGS);

Datum gist_row_acl_compress(PG_FUNCTION_ARGS);
Datum gist_row_acl_decompress(PG_FUNCTION_ARGS);
Datum gist_row_acl_same(PG_FUNCTION_ARGS);
Datum gist_row_acl_penalty(PG_FUNCTION_ARGS);
Datum gist_row_acl_union(PG_FUNCTION_ARGS);
Datum gist_row_acl_consistent(PG_FUNCTION_ARGS);
Datum gist_row_acl_picksplit(PG_FUNCTION_ARGS);

/* Local text to acl convertion function with error handling */
static row_acl
_dat_to_acl(row_acl p_acl, const char* p_str, size_t p_len)
{
    int error = 0;

    p_acl = row_acl_from_dat(p_acl, p_str, p_len, &error);

    if(error != 0){
        p_acl = 0;

        /* FIXME p_str is not guaranteed to be '\0' terminated. Error message
         * may contain junk on end. */
#if 0
        ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
            errmsg("incorrect access mode: '%s'", p_str)));
#else
        if(p_len < 384){
            char    str_buf[384];

            memcpy(str_buf, p_str, p_len);
            str_buf[p_len] = '\0';

            ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
                errmsg("incorrect access mode: '%s'", str_buf)));
        } else {
            ereport(ERROR, (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
                errmsg("incorrect access mode")));
        };
#endif
    };

    return p_acl;
}

static inline row_acl
_str_to_acl(row_acl p_acl, const char* p_str)
{
    return _dat_to_acl(p_acl, p_str, -1);
}

/*
 * text input/output
 */

/* @param cstring
 * @return row_acl */
Datum
row_acl_in(PG_FUNCTION_ARGS)
{
    ROW_ACL_RETURN(_str_to_acl(ROW_ACL_NOTHING, PG_GETARG_CSTRING(0)));
}

/* @param row_acl
 * @return cstring */
Datum
row_acl_out(PG_FUNCTION_ARGS)
{
    PG_RETURN_POINTER(row_acl_to_str(ROW_ACL_GETARG(0),
        palloc(ROW_ACL_STR_LENGTH)));
}

/*
 * operator impl functions
 */

/* @param row_acl
 * @param row_acl
 * @return boolean */
Datum
row_acl_eq(PG_FUNCTION_ARGS)
{
    PG_RETURN_BOOL(ROW_ACL_GETARG(0) == ROW_ACL_GETARG(1));
}

/* @param row_acl
 * @param row_acl
 * @return boolean */
Datum
row_acl_neq(PG_FUNCTION_ARGS)
{
    PG_RETURN_BOOL(ROW_ACL_GETARG(0) != ROW_ACL_GETARG(1));
}

/* @param row_acl
 * @param row_acl
 * @return boolean */
Datum
row_acl_is_intersects(PG_FUNCTION_ARGS)
{
    PG_RETURN_BOOL((ROW_ACL_GETARG(0) & ROW_ACL_GETARG(1)) != 0);
}

/* @param row_acl
 * @param row_acl
 * @return boolean */
Datum
row_acl_contains(PG_FUNCTION_ARGS)
{
    const row_acl b = ROW_ACL_GETARG(1);

    PG_RETURN_BOOL((ROW_ACL_GETARG(0) & b) == b);
}

/* @param row_acl
 * @param row_acl
 * @return boolean */
Datum
row_acl_within(PG_FUNCTION_ARGS)
{
    const row_acl a = ROW_ACL_GETARG(0);

    PG_RETURN_BOOL((a & ROW_ACL_GETARG(1)) == a);
}

/*
 * int convertion
 */

/* @param integer
 * @return row_acl */
Datum
int_to_row_acl(PG_FUNCTION_ARGS)
{
    ROW_ACL_RETURN(PG_GETARG_INT32(0));
}

/* @param row_acl
 * @return integer */
Datum
row_acl_to_int(PG_FUNCTION_ARGS)
{
    PG_RETURN_INT32(ROW_ACL_GETARG(0));
}

/*
 * text convertion
 */

/* @param text
 * @return row_acl */
Datum
text_to_row_acl(PG_FUNCTION_ARGS)
{
    return DirectFunctionCall1(row_acl_set_alt, PG_GETARG_DATUM(0));
}

/* @param row_acl
 * @return text */
Datum
row_acl_to_text(PG_FUNCTION_ARGS)
{
    text*   result;
    char    buffer[ROW_ACL_STR_LENGTH];

    result = palloc(VARHDRSZ + ROW_ACL_STR_LENGTH);

    memcpy(VARDATA(result), row_acl_to_str(ROW_ACL_GETARG(0), buffer),
        ROW_ACL_STR_LENGTH - 1);
    VARATT_SIZEP(result) = VARHDRSZ + ROW_ACL_STR_LENGTH - 1;

    PG_RETURN_TEXT_P(result);
}

/* @param row_acl
 * @param text
 * @return row_acl */
Datum
row_acl_set(PG_FUNCTION_ARGS)
{
    const text* t = PG_GETARG_TEXT_P(1);

    ROW_ACL_RETURN(_dat_to_acl(ROW_ACL_GETARG(0),
        VARDATA(t), VARSIZE(t) - VARHDRSZ));
}

/* @param text
 * @return row_acl */
Datum
row_acl_set_alt(PG_FUNCTION_ARGS)
{
    return DirectFunctionCall2(row_acl_set, ROW_ACL_TO_DATUM(ROW_ACL_NOTHING),
        PG_GETARG_DATUM(0));
}

/*
 * operations
 */

/* @param row_acl
 * @param row_acl
 * @return row_acl */
Datum
row_acl_intersect(PG_FUNCTION_ARGS)
{
    ROW_ACL_RETURN(ROW_ACL_GETARG(0) & ROW_ACL_GETARG(1));
}

/* @param row_acl
 * @param row_acl
 * @return row_acl */
Datum
row_acl_union(PG_FUNCTION_ARGS)
{
    ROW_ACL_RETURN(ROW_ACL_GETARG(0) | ROW_ACL_GETARG(1));
}

/* @param row_acl
 * @param row_acl
 * @return row_acl */
Datum
row_acl_subtract(PG_FUNCTION_ARGS)
{
    ROW_ACL_RETURN(ROW_ACL_GETARG(0) ^ ROW_ACL_GETARG(1));
}

/*
 * gist functions
 */

/* @param internal
 * @param row_acl
 * @param integer
 * @return boolean
 *
 * Given a predicate p on a tree page, and a user query, q, this method will
 * return false if it is certain that both p and q cannot be true for a given
 * data item.
 */
Datum
gist_row_acl_consistent(PG_FUNCTION_ARGS)
{
    const GISTENTRY*    entry;
    row_acl             entry_key;
    row_acl             query;
    StrategyNumber      strategy;
    bool                result;

    entry = (const GISTENTRY*)PG_GETARG_POINTER(0);
    query = ROW_ACL_GETARG(1);
    strategy = PG_GETARG_UINT16(2);
 
    entry_key = ROW_ACL_FROM_DATUM(entry->key);
    result = false;

    switch(strategy){
        case RTOverlapStrategyNumber :      /* && */
            result = ((entry_key & query) != ROW_ACL_NOTHING);
            break;

        case RTSameStrategyNumber :         /* = */
            if(GIST_LEAF(entry)){
                result = (entry_key == query);
            } else {
                result = ((entry_key & query) == query);
            };
            break;

        case RTContainsStrategyNumber :     /* &> */
            result = ((entry_key & query) == query);
            break;

        case RTContainedByStrategyNumber :  /* &< */
            if(GIST_LEAF(entry)){
                result = ((entry_key & query) == entry_key);
            } else {
                result = ((entry_key & query) != ROW_ACL_NOTHING);
            };
    };

    PG_RETURN_BOOL(result);
}

/* @param internal
 * @param internal
 * @return row_acl
 *
 * This method consolidates information in the tree. Given a set of entries,
 * this function generates a new predicate that is true for all the entries.
 */
Datum
gist_row_acl_union(PG_FUNCTION_ARGS)
{
    const GistEntryVector*  entry_vec;
    int*                    size;
    int                     i;
    int                     n;
    row_acl                 u;

    entry_vec = (const GistEntryVector*)PG_GETARG_POINTER(0);
    size = (int*)PG_GETARG_POINTER(1);

    n = entry_vec->n;
    u = ROW_ACL_NOTHING;

    for(i = 0; i < n; ++i){
        u |= ROW_ACL_FROM_DATUM(entry_vec->vector[i].key);
    };

    *size = sizeof(row_acl);

    ROW_ACL_RETURN(u);
}

/* @param internal
 * @return internal
 *
 * Converts the data item into a format suitable for physical storage in an
 * index page.
 */
Datum
gist_row_acl_compress(PG_FUNCTION_ARGS)
{
    PG_RETURN_DATUM(PG_GETARG_DATUM(0));
}

/* @param internal
 * @return internal
 *
 * The reverse of the compress method. Converts the index representation of
 * the data item into a format that can be manipulated by the database.
 */
Datum
gist_row_acl_decompress(PG_FUNCTION_ARGS)
{
    PG_RETURN_DATUM(PG_GETARG_DATUM(0));
}

/* @param internal
 * @param internal
 * @param internal
 * @return internal
 *
 * Returns a value indicating the "cost" of inserting the new entry into a
 * particular branch of the tree. items will be inserted down the path of
 * least penalty in the tree.
 */
Datum
gist_row_acl_penalty(PG_FUNCTION_ARGS)\
{
    const row_acl o = ROW_ACL_FROM_DATUM( ((const GISTENTRY*)PG_GETARG_POINTER(0))->key );
    const row_acl n = ROW_ACL_FROM_DATUM( ((const GISTENTRY*)PG_GETARG_POINTER(1))->key );
    float*          result = (float*)PG_GETARG_POINTER(2);

    *result = (float)(row_acl_count_ones(o | n) - row_acl_count_ones(o));

    PG_RETURN_POINTER(result);
}

/* @param internal
 * @param internal
 * @return internal
 *
 * When a page split is necessary, this function decides which entries on the
 * page are to stay on the old page, and which are to move to the new page.
 */
Datum
gist_row_acl_picksplit(PG_FUNCTION_ARGS)
{
    const GistEntryVector*  entry_vec;
    GIST_SPLITVEC*          split_vec;
    OffsetNumber            i;
    OffsetNumber            max_offset;
    int                     offs_sizes;

    entry_vec = (const GistEntryVector*)PG_GETARG_POINTER(0);
    split_vec = (GIST_SPLITVEC*)PG_GETARG_POINTER(1);

    max_offset = entry_vec->n - 1;
    offs_sizes = sizeof(OffsetNumber) * (max_offset + 2);

    split_vec->spl_left = (OffsetNumber*)palloc(offs_sizes);
    split_vec->spl_right = (OffsetNumber*)palloc(offs_sizes);
    split_vec->spl_ldatum = ROW_ACL_TO_DATUM(ROW_ACL_NOTHING);
    split_vec->spl_rdatum = ROW_ACL_TO_DATUM(ROW_ACL_NOTHING);
    split_vec->spl_nleft = 0;
    split_vec->spl_nright = 0;

    for(i = FirstOffsetNumber; i <= max_offset; i = OffsetNumberNext(i)){
        if((i % 2) == 0){
            split_vec->spl_left[split_vec->spl_nleft++] = i;
            split_vec->spl_ldatum =
                ROW_ACL_TO_DATUM(ROW_ACL_FROM_DATUM(split_vec->spl_ldatum) |
                                 ROW_ACL_FROM_DATUM(entry_vec->vector[i].key));
        } else {
            split_vec->spl_right[split_vec->spl_nright++] = i;
            split_vec->spl_rdatum =
                ROW_ACL_TO_DATUM(ROW_ACL_FROM_DATUM(split_vec->spl_rdatum) |
                                 ROW_ACL_FROM_DATUM(entry_vec->vector[i].key));
        };
    };

    PG_RETURN_POINTER(split_vec);
}

/* @param row_acl
 * @param row_acl
 * @param internal
 * @return internal
 *
 * Returns true if two entries are identical, false otherwise.
 */
Datum
gist_row_acl_same(PG_FUNCTION_ARGS)
{
    bool*   result;

    result = (bool*)PG_GETARG_POINTER(2);
    *result = (ROW_ACL_GETARG(0) == ROW_ACL_GETARG(1));

    PG_RETURN_POINTER(result);
}
