/* -*- C -*-
 * File: "/home/arn/proj/cms1/db/row_acl/row_acl_selec.c"
 * Author: "Anton Yabchinskiy, <arn@antora.ru>"
 * Copyright: ""
 * Created: "Wed, 11 Jul 2007 17:25:40 +0400 (arn@doriath.my.domain)"
 * Last modified: "Tue, 17 Jul 2007 15:55:41 +0400 (arn@doriath.my.domain)"
 * vim:set sts=4 ts=4 sw=4 ft=c:
 * $Id: row_acl_selec.c,v 1.3 2007/07/17 11:55:35 anton Exp $
 */

#include <postgres.h>
#include <fmgr.h>
#include <utils/selfuncs.h>
#include <utils/lsyscache.h>
#include <utils/syscache.h>
#include <catalog/pg_statistic.h>
#include <nodes/memnodes.h>

#include "row_acl.h"
#include "row_acl_debug.h"
#include "row_acl_stats.h"
#include "row_acl_text.h"

/*
 * pg infos
 */
PG_FUNCTION_INFO_V1(row_acl_restrict_sel);
PG_FUNCTION_INFO_V1(row_acl_join_sel);

/* 
 * prototypes
 */
Datum row_acl_restrict_sel(PG_FUNCTION_ARGS);
Datum row_acl_join_sel(PG_FUNCTION_ARGS);

/*
 * restrict
 *
 * Works only for && operator for now.
 * Should check for operator somehow and
 * use other estimation methods for &> and &<
 */
Datum
row_acl_restrict_sel(PG_FUNCTION_ARGS)
{
    PlannerInfo*        root;
    Oid                 operator;
    List*               arg_list;
    int32               var_relid;

    Node*               other;
    VariableStatData    var_data;
    bool                var_on_left;

    float8              selec;

    /* init */
    root = (PlannerInfo*)PG_GETARG_POINTER(0);
    operator = PG_GETARG_OID(1);
    arg_list = (List*)PG_GETARG_POINTER(2);
    var_relid = PG_GETARG_INT32(3);
    selec = DEFAULT_EQ_SEL;

    ROW_ACL_DEBUG_PRINT("root { type %i, parse '%s', total_table_pages %f, "
        "tuple_fraction %f, hasJoinRTEs %i, hasOuterJoins %i, "
        "hasHavingQual %i, hasPseudoConstantQuals %i }",
        root->type, nodeToString(root->parse), root->total_table_pages,
        root->tuple_fraction, root->hasJoinRTEs, root->hasOuterJoins,
        root->hasHavingQual, root->hasPseudoConstantQuals);

    ROW_ACL_DEBUG_BLOCK(

        FmgrInfo    opinfo;
        char*       opname;

        fmgr_info(get_opcode(operator), &opinfo);
        opname = get_opname(operator);

        ROW_ACL_DEBUG_PRINT("operator %i, name '%s', opinfo { _oid %i, "
            "_nargs %i, _strict %i, _retset %i, _mcxt->name '%s' }",
            operator, opname, opinfo.fn_oid, opinfo.fn_nargs,
            opinfo.fn_strict, opinfo.fn_retset, opinfo.fn_mcxt->name);

        pfree(opname);

    );

    ROW_ACL_DEBUG_PRINT("arg_list '%s'", nodeToString(arg_list));
    ROW_ACL_DEBUG_PRINT("var_relid %i, name '%s'", var_relid,
        get_rel_name(var_relid));

    /* get variable info */
    if(get_restriction_variable(root, arg_list, var_relid,
        &var_data, &other, &var_on_left) != true)
    {
        ROW_ACL_DEBUG_PRINT("unable to get restriction variable, "
            "returning default selec of %f", DEFAULT_EQ_SEL);

        PG_RETURN_FLOAT8(DEFAULT_EQ_SEL);
    } else {
        ROW_ACL_DEBUG_PRINT("var_data { var '%s', rel '%s', vartype %i, "
            "atttype %i, atttypmod %i, isunique %i }", nodeToString(var_data.var),
            nodeToString(var_data.rel), var_data.vartype, var_data.atttype,
            var_data.atttypmod, var_data.isunique);
        ROW_ACL_DEBUG_PRINT("other '%s'", nodeToString(other));
        ROW_ACL_DEBUG_PRINT("var_on_left %i", var_on_left);
    };

    /* examine pg_statistic tuple */
    if(HeapTupleIsValid(var_data.statsTuple)){
        Form_pg_statistic   stats;

        stats = (Form_pg_statistic)GETSTRUCT(var_data.statsTuple);

        ROW_ACL_DEBUG_PRINT("statsTuple { _relid %i, _attnum %i, _nullfrac %f, "
            "_width %i, _distinct %f, _kind1 %i, _op1 %i }", stats->starelid,
            stats->staattnum, stats->stanullfrac, stats->stawidth,
            stats->stadistinct, stats->stakind1, stats->staop1);

        if(IsA(other, Const)){
            const Const*    other_const = (const Const*)other;
            row_acl         other_acl;
            float4*         numbers;
            int             numnumbers;

            ROW_ACL_DEBUG_PRINT("other_const { _type %i, _len %i, _value %#x, "
                "_isnull %i, _byval %i }", other_const->consttype, other_const->constlen,
                DatumGetUInt32(other_const->constvalue), other_const->constisnull,
                other_const->constbyval);

            other_acl = ROW_ACL_FROM_DATUM(other_const->constvalue);

            ROW_ACL_DEBUG_BLOCK(
                char    buf[32];
                ROW_ACL_DEBUG_PRINT("other_acl '%s', row_acl_freqs_intersection %f",
                    row_acl_to_str(other_acl, buf),
                    row_acl_freqs_intersection(stats->stanumbers1, other_acl));
            );

            if(get_attstatsslot(var_data.statsTuple, var_data.atttype,
                var_data.atttypmod, ROW_ACL_STATISTICS_KIND, InvalidOid,
                NULL, NULL, &numbers, &numnumbers))
            {
                ROW_ACL_DEBUG_PRINT("stats slot fetched, numnumbers %i",
                    numnumbers);

                selec = row_acl_freqs_intersection(numbers, other_acl);

                free_attstatsslot(var_data.atttype, NULL, 0, numbers, numnumbers);
            } else {
                ROW_ACL_DEBUG_PRINT("%s", "unable to get attrib stats slot");
                selec = -1;
            };
        } else {
            double  ndist;

            ROW_ACL_DEBUG_PRINT("%s", "other node is not a const");

            selec = 1.0 - stats->stanullfrac;
            ndist = get_variable_numdistinct(&var_data);

            ROW_ACL_DEBUG_PRINT("selec %f, ndist %f", selec, ndist);

            if(ndist > 1){
                selec /= ndist;
            };
        };
    } else {
        ROW_ACL_DEBUG_PRINT("%s", "var_data.statsTuple is not valid");
        selec = -1;
    };

    if(selec < 0){
        double  ndist;

        ROW_ACL_DEBUG_PRINT("%s", "selectivity cannot be estimated, "
            "using (1 / num_distinct)");

        if((ndist = get_variable_numdistinct(&var_data)) > 0){
            selec = 1.0 / ndist;
        };
    };

    /* fini */
    ReleaseVariableStats(var_data);
    CLAMP_PROBABILITY(selec);

    ROW_ACL_DEBUG_PRINT("estimated selectivity %f", selec);

    PG_RETURN_FLOAT8(selec);
}

/*
 * join
 */
Datum
row_acl_join_sel(PG_FUNCTION_ARGS)
{
    ROW_ACL_DEBUG_PRINT("returning %f", 0.001);

    PG_RETURN_FLOAT8(0.001);
}
