/* -*- C -*-
 * File: "/home/arn/proj/cms1/db/row_acl/row_acl_stats.c"
 * Author: "Anton Yabchinskiy, <arn@antora.ru>"
 * Copyright: ""
 * Created: "Wed, 11 Jul 2007 15:53:37 +0400 (arn@doriath.my.domain)"
 * Last modified: "Tue, 17 Jul 2007 15:56:32 +0400 (arn@doriath.my.domain)"
 * vim:set sts=4 ts=4 sw=4 ft=c:
 * $Id: row_acl_stats.c,v 1.3 2007/07/17 11:55:35 anton Exp $
 */

#include <postgres.h>
#include <fmgr.h>
#include <commands/vacuum.h>
#include <nodes/memnodes.h>

#include "row_acl_debug.h"
#include "row_acl_field.h"
#include "row_acl_stats.h"
#include "row_acl_text.h"

#define _DEFAULT_ROWS_TO_ANALYZE    64

/*
 * pg infos
 */
PG_FUNCTION_INFO_V1(row_acl_analyze);

/*
 * prototypes
 */
Datum row_acl_analyze(PG_FUNCTION_ARGS);

/*
 * local
 */
static void
row_acl_compute_stats(VacAttrStatsP p_stats, AnalyzeAttrFetchFunc p_fetchfunc,
    int p_samplerows, double p_totalrows)
{
    int             i;
    unsigned int    null_count;
    unsigned int    not_null_count;

    row_acl_field_p field;
    row_acl_freqs_p freqs;

    ROW_ACL_DEBUG_PRINT("p_stats->anl_context->name '%s'", p_stats->anl_context->name);

    ROW_ACL_DEBUG_PRINT("p_stats->attr-> { _relid %i, _name '%s', _typid %i, "
        "_stattarget %i, _len %i, _num %i, _ndims %i, _typmod %i, _align '%c'. "
        "_notnull %i, _hasdef %i, _islocal %i, _inhcount %i }",
        p_stats->attr->attrelid, NameStr(p_stats->attr->attname),
        p_stats->attr->atttypid, p_stats->attr->attstattarget,
        p_stats->attr->attlen, p_stats->attr->attnum, p_stats->attr->attndims,
        p_stats->attr->atttypmod, p_stats->attr->attalign,
        p_stats->attr->attnotnull, p_stats->attr->atthasdef,
        p_stats->attr->attislocal, p_stats->attr->attinhcount);

    ROW_ACL_DEBUG_PRINT("p_stats->attrtype-> { _name '%s', _namespace %i, "
        "_owner %i, _len %i, _byval %i, _type '%c', _isdefined %i, "
        "_delim '%c', _relid %i, _elem %i, _input %i, _output %i, _receive %i, "
        "_send %i, _analyze %i, _align '%c', _storage '%c', _notnull %i, "
        "_basetype %i, _typmod %i, _ndims %i, _defaultbin '%s', _default '%s' }",
        NameStr(p_stats->attrtype->typname), p_stats->attrtype->typnamespace,
        p_stats->attrtype->typowner, p_stats->attrtype->typlen,
        p_stats->attrtype->typbyval, p_stats->attrtype->typtype,
        p_stats->attrtype->typisdefined, p_stats->attrtype->typdelim,
        p_stats->attrtype->typrelid, p_stats->attrtype->typelem,
        p_stats->attrtype->typinput, p_stats->attrtype->typoutput,
        p_stats->attrtype->typreceive, p_stats->attrtype->typsend,
        p_stats->attrtype->typanalyze, p_stats->attrtype->typalign,
        p_stats->attrtype->typstorage, p_stats->attrtype->typnotnull,
        p_stats->attrtype->typbasetype, p_stats->attrtype->typtypmod,
        p_stats->attrtype->typndims, p_stats->attrtype->typdefaultbin.vl_dat,
        p_stats->attrtype->typdefault.vl_dat);

    ROW_ACL_DEBUG_PRINT("p_samplerows %i, p_totalrows %f", p_samplerows, p_totalrows);

    /* init */
    null_count = 0;
    not_null_count = 0;
    field = MemoryContextAllocZero(p_stats->anl_context, sizeof(row_acl_field_t));
    freqs = MemoryContextAllocZero(p_stats->anl_context, sizeof(row_acl_freqs_t));

    /* collect stats */
    for(i = 0; i < p_samplerows; ++i){
        bool    is_null;
        row_acl acl;

        acl = ROW_ACL_FROM_DATUM(p_fetchfunc(p_stats, i, &is_null));

        ROW_ACL_DEBUG_PRINT("i %i, is_null %i", i, is_null);

        if(is_null != true){
            ++not_null_count;

            ROW_ACL_DEBUG_PRINT("acl 0x%04x is %sdistinct",
                acl, row_acl_field_get(field, acl) ? "not " : "");

            row_acl_field_set(field, acl);
            row_acl_freqs_update(freqs, acl);
        } else {
            ++null_count;
        };
    };

    ROW_ACL_DEBUG_PRINT("null_count %u, not_null_count %u, distinct %i",
        null_count, not_null_count, row_acl_field_uniqs(field));

    /* fill statistics */
    if(not_null_count > 0){
        row_acl_freqs_normalise(freqs, p_samplerows);

        ROW_ACL_DEBUG_BLOCK(

            char    buf[32];

            ROW_ACL_DEBUG_PRINT("%s", "normalised bit frequencies:");
            for(i = 0; i < ROW_ACL_BIT_LENGTH; ++i){
                ROW_ACL_DEBUG_PRINT("'%s' %f", row_acl_to_str(1 << i, buf),
                    freqs[i]);
            };

        );

        /* fill stats struct */
        p_stats->stanullfrac = (double)null_count / (double)p_samplerows;
        p_stats->stawidth = sizeof(row_acl);
        p_stats->stadistinct = (double)row_acl_field_uniqs(field) / (double)not_null_count;

        p_stats->stakind[0] = ROW_ACL_STATISTICS_KIND;
        p_stats->staop[0] = InvalidOid;
        p_stats->numnumbers[0] = ROW_ACL_BIT_LENGTH;
        p_stats->stanumbers[0] = freqs;

        p_stats->stats_valid = true;

        ROW_ACL_DEBUG_PRINT("p_stats { _valid %i, _nullfrac %f, _width %i, "
            "_distinct %f, _kind[0] %i, _op[0] %i, _nnums[0] %i, _nvals[0] %i }",
            p_stats->stats_valid, p_stats->stanullfrac, p_stats->stawidth,
            p_stats->stadistinct, p_stats->stakind[0], p_stats->staop[0],
            p_stats->numnumbers[0], p_stats->numvalues[0]);
    } else {
        ROW_ACL_DEBUG_PRINT("%s", "all rows are null, no useful statistics "
            "can be built");

        p_stats->stats_valid = false;
    };

    pfree(field);
}

/*
 * analyze
 */
Datum
row_acl_analyze(PG_FUNCTION_ARGS)
{
    VacAttrStats*       input;
    Form_pg_attribute   attrs;

    input = (VacAttrStats*)PG_GETARG_POINTER(0);
    attrs = input->attr;

    input->compute_stats = row_acl_compute_stats;
    input->extra_data = 0;
    input->minrows = attrs->attstattarget < 0 ?
        _DEFAULT_ROWS_TO_ANALYZE : 305 * attrs->attstattarget;

    ROW_ACL_DEBUG_PRINT("minrows %i, anl_context->name '%s'",
        input->minrows, input->anl_context->name);

    PG_RETURN_BOOL(true);
}
